; 232-485 Transciever

;
; define PIC device type
;
;	device	pic16f84

;
; define config fuses
;
;	config	CP=off,WDT=off,PWRT=off,OSC=hs

	__config _wdt_off & _hs_osc
	
;
; include PIC register definitions, and some macros
;
	include "../include/p16f84.inc"
	include "../include/picmac_gpasm.h"


;
;                        -----------
;                       |     U     |
;            SPI CLK  --|RA2     RA1|-- 
;            SPI DIN  --|RA3     RA0|-- 
;           SPI DOUT  --|RA4        |
;                       |           | 
;                       |           |
;                       |           |
;             232_CS  --|RB0     RB7|-- 
;             485_CS  --|RB1     RB6|-- 
;             232_IRQ --|RB2     RB5|-- 
;             485_IRQ --|RB3     RB4|-- 485 RE/DE
;                       |           |
;                        -----------





; The following are the configuration bits that can be set





























;
; bit definitions for two LEDs connected to port A bits 0 and 1
; bit masks can be computed from bit numbers with the left shift operator.
;

; this ordering is good for box 1
















select_232_uart macro
	movlw	PORTB 
	movwf	uart_3110_port
	movlw	1 
	movwf	uart_3110_mask
	endm
	
select_485_uart	macro
	movlw	PORTB 
	movwf	uart_3110_port
	movlw	2 
	movwf	uart_3110_mask
	endm

;
; define some register file variables
;

	SEGU data
	org	0x0c

	SEG code

;
; code start
;
	org	0
	goto start

	org 4
	clrf	TMR0
	bcf		INTCON,T0IF
	bcf		INTCON,GIE
	bcf		PORTB,4 
	retfie
	
	
start

	clrf	PORTA
	clrf	PORTB
	
	bsf		PORTB,0 
	bsf		PORTB,1 
	

	bsf	STATUS,RP0			;register page 1
	movlw	B'00010000'			;all others output
	movwf	TRISA				;other PORTA pins as inputs
	movlw	B'00001100'			;b7,b6 input
	movwf	TRISB

	movlw	B'11010011'		; select 1:16 timer prescale
	movwf	OPTION_REG

	bcf	STATUS,RP0			;register page 0

	clrf	INTCON

	bsf	INTCON,T0IE		; turn on timer overflow interrupt
	bcf	INTCON,INTE		; turn off RB0 edge interrupt
	bcf	INTCON,RBIE

; globally enable interrupts
	bsf	INTCON,GIE

	select_232_uart
	
	movlw	HIGH(0xc000  | 0x0400  | 0x09  | 0x0020  )
	movwf	uart_3110_parmhi
	movlw	LOW(0xc000  | 0x0400  | 0x09  | 0x0020  )
	movwf	uart_3110_parmlo
	call	UART_232_INIT_PARM
	
	select_485_uart
	
	movlw	HIGH(0xc000  | 0x0400  | 0x09  | 0x0020  )
	movwf	uart_3110_parmhi
	movlw	LOW(0xc000  | 0x0400  | 0x09  | 0x0020  )
	movwf	uart_3110_parmlo
	call	UART_232_INIT_PARM


;testloop
	select_232_uart
	movlw	'J'
	call	UART_232_TX
	movlw	'P'
	call	UART_232_TX
	movlw	'0'
	call	UART_232_TX
;	
;	movlw	0xff
;	call	delay
;
;	goto testloop
	
main_loop		; essentially spin on either IRQ... if one is asserted, process it
	btfss	PORTB,2 
	goto	handle_232_irq
	

	btfss	PORTB,3 
	goto	handle_485_irq

	
	goto	main_loop

; need to properly handle the parity bit in here... just copy it, but still

handle_232_irq
	; we need to read in a byte from the 232 uart, then write it out to the 485 uart
	; a room for future improvement would detect if the 485 was currently receiving
	; before sending, since this could result in a bus conflict
	
	select_232_uart
	call	UART_232_RX
	
	movf	spi_high,W
	andlw	1
	movwf	uart_3110_parity

	
	
	select_485_uart
	movf	spi_low,W
	call	UART_232_TX

	bsf		PORTB,4 

	clrf	TMR0			; reset the timer
	bcf		INTCON,T0IF
	bsf		INTCON,GIE		; enable interrupt, when timer period passes with
							; no transmissions, then bus is relinquished


;	select_232_uart
;	movlw	'2'
;	call	UART_232_TX

	; hopefully the last bit will have been transmitted while were checking the
	; 232 interface...  At faster baudrates this becomes less of a problem, as the
	; UART_TX procedure occupies more and more of a baud period


	goto	main_loop

handle_485_irq
	select_485_uart
	call	UART_232_RX
	
	select_232_uart
	movf	spi_low,W
	call	UART_232_TX

;	movlw	'4'
;	call	UART_232_TX
	
	goto	main_loop
	






; This file defines routines for communicating over a SPI bus.
;
; PORTA,2 , PORTA,3 , and PORTA,4  must be predefined

	SEGU data
	
spi_low DS 1
spi_high DS 1
spi_temp DS 1
spi_temp2 DS 1
spi_cnt1 DS 1
spi_cnt2 DS 1
	
	SEG code	

;------------
; SPI_READ16
;
; Reads a 16 bit number from the SPI bus
;
; 4+8*8+4+8*8+2=138

SPI_READ16   
    movlw   8
    movwf   spi_cnt1
    clrf    spi_high
    bcf STATUS,C
spi_read_loop  
    bsf PORTA,2 
    btfsc   PORTA,4 
    bsf STATUS,C  
    rlf spi_high,F

    bcf PORTA,2 

    decfsz  spi_cnt1,F 
    goto    spi_read_loop

    movlw   8
    movwf   spi_cnt2
    clrf    spi_low
    bcf STATUS,C
spi_read_loop2        
    bsf PORTA,2        
    btfsc   PORTA,4   
    bsf STATUS,C
    rlf spi_low,F
    bcf PORTA,2        
    decfsz  spi_cnt2,F
    goto    spi_read_loop2
    
    return

;------------
; SPI_READ12
;
; Reads a 12 bit number from the SPI bus
;
; 4+8*4+4+8*8+2=106

SPI_READ12   
    movlw   4
    movwf   spi_cnt1
    clrf    spi_high
    bcf STATUS,C
spi_read12_loop  
    bsf PORTA,2 
    bcf PORTA,2 
    btfsc   PORTA,4 
    bsf STATUS,C  
    rlf spi_high,F
    decfsz  spi_cnt1,F 
    goto    spi_read12_loop

    movlw   8
    movwf   spi_cnt2
    clrf    spi_low
    bcf STATUS,C
spi_read12_loop2        
    bsf PORTA,2        
    bcf PORTA,2        
    btfsc   PORTA,4   
    bsf STATUS,C
    rlf spi_low,F
    decfsz  spi_cnt2,F
    goto    spi_read12_loop2
    
    return

	
;-------------
; SPI_READWRITE16
;
; spi_high and spi_low have the data to be sent
; and they are filled with the received data
;
; 12 + SPI_READWRITE8*2= 12+109*2=230

SPI_READWRITE16
	movf	spi_high,W
	call	SPI_READWRITE8
	movf	spi_temp2,W
	movwf	spi_high
	
	movf	spi_low,W
	call	SPI_READWRITE8
	movf	spi_temp2,W
	movwf	spi_low
	retlw 	0
	
;-------------
; SPI_READWRITE8
;
; W has the data to be written
; spi_temp2 has the received data
;
; they will be overwitten with the received data
;
; 3+13*8+2=109


SPI_READWRITE8
	movwf	spi_temp

	movlw	8
	movwf	spi_cnt1
spi_rwloop1
	rlf	spi_temp,F
	btfsc	STATUS,C
	bsf		PORTA,3 
	btfss	STATUS,C
	bcf		PORTA,3 	
	
	bsf		PORTA,2 
	
	bcf		STATUS,C
	btfsc	PORTA,4 
	bsf		STATUS,C
	rlf		spi_temp2,F
	
	bcf		PORTA,2 
	decfsz	spi_cnt1,F
	goto	spi_rwloop1
	
	retlw 0	




; Interface to MAXIM 3110 UART using my SPI interface
;
; the CS pin is assumed to be at the mem location uart_3110_port
; and a bit mask is at uart_3110_mask



; The following are the configuration bits that can be set





	SEGU data
	
uart_3110_port DS 1
uart_3110_mask DS 1
uart_3110_temp DS 1	
uart_3110_temp2 DS 1
uart_3110_parmhi DS 1
uart_3110_parmlo DS 1

uart_rx_buf DS 1
uart_3110_parity DS 1
	
	SEG code

;----------------
; UART_INIT
;
; Called to initialize the UART with 
; a default configuration
;

UART_232_INIT

	movlw	HIGH(0xc000  | 0x09   )
	movwf	uart_3110_parmhi
	movlw	LOW(0xc000  | 0x09   )
	movwf	uart_3110_parmlo
	call	UART_232_INIT_PARM
	
	return
	

;----------------
; UART_INIT_PARM
;
; Called to initialize the UART with a passed config register
; in uart_3110_temp and uart_3110_temp2
;
UART_232_INIT_PARM

	movf	uart_3110_port,W
	movwf	FSR

	movf	uart_3110_parmhi,W
	movwf	spi_high
	movf	uart_3110_parmlo,W
	movwf	spi_low
	
	comf	uart_3110_mask,W	; load an inverted (negative) mask into W
	andwf	IND0,F				; clear the masked bit
	call	SPI_READWRITE16
	movf	uart_3110_mask,W	; load the normal mask into W
	iorwf	IND0,F				; set the CS bit
	
	movlw	HIGH(0x4000 )
	movwf	spi_high
	clrf	spi_low

	comf	uart_3110_mask,W	; load an inverted (negative) mask into W
	andwf	IND0,F				; clear the masked bit
	call	SPI_READWRITE16
	movf	uart_3110_mask,W	; load the normal mask into W
	iorwf	IND0,F				; set the CS bit
	
	movf	spi_low,W
	subwf	uart_3110_parmlo,W
	btfss	STATUS,Z
	goto	UART_232_INIT_PARM
	
	movlw	B'00111111'
	andwf	spi_high,F
	movf	uart_3110_parmhi,W
	andlw	B'00111111'
	subwf	spi_high,W
	btfss	STATUS,Z
	goto	UART_232_INIT_PARM
	
	retlw 0


;----------------
; UART_RX
;
; receives a byte into spi_low
; if no bytes are available 0 is returned in W otherwise
; 1 is returned...
;
; 6(7)+SPI_READ16=6(7)+138=144(145)
;

UART_232_RX
	bcf		PORTA,3 

	movf	uart_3110_port,W
	movwf	FSR

	comf	uart_3110_mask,W
	andwf	IND0,F
	call	SPI_READ16
	movf	uart_3110_mask,W
	iorwf	IND0,F

	btfss	spi_high,7
	retlw	0
	retlw	1


;----------------
; UART_232_TX_WAIT
;
; wait until a byte can be sent
UART_232_TX_WAIT
	movf	uart_3110_port,W
	movwf	FSR
	
u2tw_loop
	movlw	HIGH(0x4000 )
	movwf	spi_high
	clrf	spi_low
	
	comf	uart_3110_mask,W
	andwf	IND0,F
	call	SPI_READWRITE16
	movf	uart_3110_mask,W
	iorwf	IND0,F
	
	btfss	spi_high,6
	goto	u2tw_loop
	return

	
;----------------
; UART_TX
;
; the byte to send is in W
; and the appropriate parity bit is in uart_3110_parity
;

UART_232_TX
	; first have to wait to make sure the uart has room
	; for our character....
	movwf	uart_3110_temp

;	call	UART_232_TX_WAIT
	
	movf	uart_3110_temp,W
	movwf	spi_low
	movf	uart_3110_parity,W
	iorlw	HIGH(0x8000 )
	movwf	spi_high

	comf	uart_3110_mask,W
	andwf	IND0,F
	call	SPI_READWRITE16
	movf	uart_3110_mask,W
	iorwf	IND0,F

	retlw 0

;----------------
; UART_232_PARITY_CHECK
;
; calculates the parity of the byte in spi_low and compares it against
; the received parity bit in spi_high.  If the conditions for odd parity
; are met, W=0 on return. otherwise w=1
;

UART_232_PARITY_CHECK
	movf	spi_low,W
	movwf	uart_3110_temp
	swapf	uart_3110_temp,F
	xorwf	uart_3110_temp,W
	movwf	uart_3110_temp
	rrf		uart_3110_temp,F
	rrf		uart_3110_temp,F
	xorwf	uart_3110_temp,W
	movwf	uart_3110_temp
	rrf		uart_3110_temp,F
	xorwf	uart_3110_temp,F
	movlw	1
	andwf	uart_3110_temp,F
	andwf	spi_high,W
	subwf	uart_3110_temp,W
	btfss	STATUS,Z
	retlw	0
	retlw	1






	SEGU data

delay_cnt1 DS 1
delay_cnt2 DS 1
delay_cnt3 DS 1

	SEG code

; routines for waiting
; good for doing nothing

delay2
    movwf   delay_cnt3
delay2_loop
    movlw   0
    call    delay
    decfsz  delay_cnt3,F
    goto    delay2_loop 
    return

;
; delay subroutine
; input: delay count in W
;
; inner loop duration approx:
; 5*256+3 = 1283 cycles ->   
; 1.28ms with 4MHz crystal (1MHz instruction time)
;
delay   movwf   delay_cnt1
    clrf    delay_cnt2
delay_a nop
    nop
    incfsz  delay_cnt2,F
    goto    delay_a
    decfsz  delay_cnt1,F
    goto    delay_a
    return












	end
